#include "GeoDrawDemo.h"

struct Vertex
{
	XMFLOAT3 pos;
	XMFLOAT4 color;
};

GeoDrawDemo::GeoDrawDemo() :m_pInputLayout(nullptr), m_pVertexBuffer(nullptr),m_pFx(nullptr),
m_theta(1.5f*XM_PI), m_phi(0.4f*XM_PI), m_radius(20.0f)
{
	//ʼ任
	//
	XMMATRIX gridWorld = XMMatrixIdentity();
	XMStoreFloat4x4(&m_gridWorld, gridWorld);
	//
	//XMMatrixTranslationƽƱ任ָϵƫ
	XMMATRIX boxWorld = XMMatrixTranslation(0.f, 0.75f, 0.f);
	XMStoreFloat4x4(&m_boxWorld, boxWorld);
	//5
	XMMATRIX sphereWorld = XMMatrixTranslation(0.f, 3.5f, 0.f);
	XMStoreFloat4x4(&m_sphereWorld[4], sphereWorld);
	//4Բ4λ4ԳƵϣ
	for (UINT i = 0; i < 2; ++i)
	{
		for (UINT j = 0; j < 2; ++j)
		{
			XMMATRIX cylinderWorld = XMMatrixTranslation(-5.f + i*10.f, 1.f, -5.f + j*10.f);
			XMStoreFloat4x4(&m_cylinderWorld[i * 2 + j], cylinderWorld);
			XMMATRIX sphereWorld = XMMatrixTranslation(-5.f + i*10.f, 4.f, -5.f + j*10.f);
			XMStoreFloat4x4(&m_sphereWorld[i * 2 + j], sphereWorld);
		}
	}
}

GeoDrawDemo::~GeoDrawDemo()
{

}

bool GeoDrawDemo::LoadContent()
{
	HRESULT hr;

	//create vertex buffer	
	GeometryGenerator::GetInstance()->CreateGrid(20.f, 20.f, 50, 50, m_grid);
	GeometryGenerator::GetInstance()->CreateBox(2, 1.5f, 2, m_box);
	GeometryGenerator::GetInstance()->CreateSphere(2, 30, 30, m_sphere);
	GeometryGenerator::GetInstance()->CreateCylinder(0.5f, 0.5f, 2, 20, 20, m_cylinder);
	GeometryGenerator::GetInstance()->AddCylinderTopCap(0.5f, 0.5f, 2, 20, 20, m_cylinder);

	m_gridVStart = 0;
	m_gridIStart = 0;
	m_boxVStart = m_grid.vertices.size();
	m_boxIStart = m_grid.indices.size();
	m_sphereVStart = m_boxVStart + m_box.vertices.size();
	m_sphereIStart = m_boxIStart + m_box.indices.size();
	m_cylinderVStart = m_sphereVStart + m_sphere.vertices.size();
	m_cylinderIStart = m_sphereIStart + m_sphere.indices.size();
	
	UINT totalVertices = m_cylinderVStart + m_cylinder.vertices.size();
	UINT totalIndices = m_cylinderIStart + m_cylinder.indices.size();

	std::vector<Vertex> vertices(totalVertices);
	for (UINT i = 0; i < m_grid.vertices.size(); ++i)
	{
		vertices[m_gridVStart + i].pos = m_grid.vertices[i].pos;
		vertices[m_gridVStart + i].color = XMFLOAT4(0.f, 0.f, 0.f, 1.f);
	}
	for (UINT i = 0; i < m_box.vertices.size(); ++i)
	{
		vertices[m_boxVStart + i].pos = m_box.vertices[i].pos;
		vertices[m_boxVStart + i].color = XMFLOAT4(0.f, 0.f, 0.f, 1.f);
	}
	for (UINT i = 0; i < m_sphere.vertices.size(); ++i)
	{
		vertices[m_sphereVStart + i].pos = m_sphere.vertices[i].pos;
		vertices[m_sphereVStart + i].color = XMFLOAT4(0.f, 0.f, 0.f, 1.f);
	}
	for (UINT i = 0; i < m_cylinder.vertices.size(); ++i)
	{
		vertices[m_cylinderVStart + i].pos = m_cylinder.vertices[i].pos;
		vertices[m_cylinderVStart + i].color = XMFLOAT4(0.f, 0.f, 0.f, 1.f);
	}

	D3D11_BUFFER_DESC vertexDesc;
	ZeroMemory(&vertexDesc, sizeof(vertexDesc));
	vertexDesc.Usage = D3D11_USAGE_IMMUTABLE;
	vertexDesc.BindFlags = D3D11_BIND_VERTEX_BUFFER;
	vertexDesc.ByteWidth = sizeof(Vertex)* totalVertices;
	D3D11_SUBRESOURCE_DATA resourceData;
	ZeroMemory(&resourceData, sizeof(resourceData));
	resourceData.pSysMem = &vertices[0];
	hr = m_pd3dDevice->CreateBuffer(&vertexDesc, &resourceData, &m_pVertexBuffer);
	if (FAILED(hr))
	{
		return false;
	}
	//set vertex buffer
	UINT stride = sizeof(Vertex);
	UINT offset = 0;
	m_pImmediateContext->IASetVertexBuffers(0, 1, &m_pVertexBuffer, &stride, &offset);
	//set primitive topology
	m_pImmediateContext->IASetPrimitiveTopology(D3D11_PRIMITIVE_TOPOLOGY_TRIANGLELIST);



	//create index buffer
	std::vector<UINT> indices(totalIndices);
	for (UINT i = 0; i < m_grid.indices.size(); ++i)
	{
		indices[m_gridIStart + i] = m_grid.indices[i];
	}
	for (UINT i = 0; i < m_box.indices.size(); ++i)
	{
		indices[m_boxIStart + i] = m_box.indices[i];
	}
	for (UINT i = 0; i < m_sphere.indices.size(); ++i)
	{
		indices[m_sphereIStart + i] = m_sphere.indices[i];
	}
	for (UINT i = 0; i < m_cylinder.indices.size(); ++i)
	{
		indices[m_cylinderIStart + i] = m_cylinder.indices[i];
	}

	D3D11_BUFFER_DESC indexDesc;
	ZeroMemory(&indexDesc, sizeof(indexDesc));
	indexDesc.Usage = D3D11_USAGE_IMMUTABLE;
	indexDesc.BindFlags = D3D11_BIND_INDEX_BUFFER;
	indexDesc.ByteWidth = sizeof(UINT)* totalIndices;

	D3D11_SUBRESOURCE_DATA indexData;
	ZeroMemory(&indexData, sizeof(indexData));
	indexData.pSysMem = &indices[0];
	hr = m_pd3dDevice->CreateBuffer(&indexDesc, &indexData, &m_pIndexBuffer);
	if (FAILED(hr))
	{
		return false;
	}

	//set index buffer
	m_pImmediateContext->IASetIndexBuffer(m_pIndexBuffer, DXGI_FORMAT_R32_UINT, 0);

	//compile shader
	ID3DBlob* errorBlob;
	DWORD shaderFlags = D3DCOMPILE_ENABLE_STRICTNESS;

	
#if defined _DEBUG || defined DEBUG
	shaderFlags = D3DCOMPILE_DEBUG;
#endif

	hr = D3DX11CompileEffectFromFile(L"color.fx", nullptr, D3D_COMPILE_STANDARD_FILE_INCLUDE, shaderFlags,
		0, m_pd3dDevice, &m_pFx, &errorBlob);
	if (FAILED(hr))
	{
		MessageBox(nullptr, (LPCWSTR)errorBlob->GetBufferPointer(), L"error", MB_OK);
		return hr;
	}
	
	m_pTechnique = m_pFx->GetTechniqueByName("ColorTech");
	m_pFxWorldViewProj = m_pFx->GetVariableByName("gWorldViewProj")->AsMatrix();


	//define the input layout
	D3D11_INPUT_ELEMENT_DESC colorLayout[] =
	{
		{ "POSITION", 0, DXGI_FORMAT_R32G32B32_FLOAT, 0, 0, D3D11_INPUT_PER_VERTEX_DATA, 0 },
		{ "COLOR", 0, DXGI_FORMAT_R32G32B32A32_FLOAT, 0, 12, D3D11_INPUT_PER_VERTEX_DATA, 0 }
	};

	UINT numLayoutElements = ARRAYSIZE(colorLayout);
	D3DX11_PASS_DESC passDesc;
	m_pTechnique->GetPassByIndex(0)->GetDesc(&passDesc);

	//create the input layout
	hr = m_pd3dDevice->CreateInputLayout(colorLayout, numLayoutElements, passDesc.pIAInputSignature,
		passDesc.IAInputSignatureSize, &m_pInputLayout);
	if (FAILED(hr))
		return hr;

	//set input layout
	m_pImmediateContext->IASetInputLayout(m_pInputLayout);

	return true;
}

void GeoDrawDemo::UnLoadContent()
{
	if (m_pVertexBuffer) m_pVertexBuffer->Release();
	if (m_pIndexBuffer)	m_pIndexBuffer->Release();
	if (m_pInputLayout)	m_pInputLayout->Release();
	//if (m_pFx) m_pFx->Release();
}

void GeoDrawDemo::Update(float dt)
{
	float x = m_radius*sinf(m_phi)*cosf(m_theta);
	float z = m_radius*sinf(m_phi)*sinf(m_theta);
	float y = m_radius*cosf(m_phi);

	XMVECTOR pos = XMVectorSet(x, y, z, 1.0f);
	XMVECTOR target = XMVectorSet(0.0f, 0.0f, 0.0f, 1.0f);
	XMVECTOR up = XMVectorSet(0.0f, 1.0f, 0.0f, 0.0f);

	XMMATRIX V = XMMatrixLookAtLH(pos, target, up);
	XMStoreFloat4x4(&m_view, V);
	XMMATRIX T = XMMatrixPerspectiveFovLH(XM_PIDIV4, m_width / static_cast<float>(m_height),
		1.0f, 1000.0f);
	XMStoreFloat4x4(&m_proj, T);
}

void GeoDrawDemo::Render()
{
	if (m_pImmediateContext == 0)
		return;
	//clear render target view
	float clearColor[4] = { 0.75f, 0.75f, 0.75f, 1.0f };
	m_pImmediateContext->ClearRenderTargetView(m_pRenderTargetView, clearColor);
	
	//clear depth/stencil view
	m_pImmediateContext->ClearDepthStencilView(m_pDepthStencilView, D3D11_CLEAR_DEPTH | D3D11_CLEAR_STENCIL,
		1.0f, 0);
	//Ϊ߿ģʽ
	D3D11_RASTERIZER_DESC rsDesc;
	ZeroMemory(&rsDesc, sizeof(rsDesc));
	rsDesc.CullMode = D3D11_CULL_BACK;
	rsDesc.DepthClipEnable = true;
	//D3D11_FILL_WIREFRAME߿ģʽƣD3D11_FILL_SOLIDʵģʽ
	rsDesc.FillMode = D3D11_FILL_WIREFRAME; 
	rsDesc.FrontCounterClockwise = false;
	ID3D11RasterizerState *rsState(nullptr);
	m_pd3dDevice->CreateRasterizerState(&rsDesc, &rsState);
	m_pImmediateContext->RSSetState(rsState);

	//set constant buffer
	XMMATRIX view = XMLoadFloat4x4(&m_view);
	XMMATRIX proj = XMLoadFloat4x4(&m_proj);


	D3DX11_TECHNIQUE_DESC techDesc;
	m_pTechnique->GetDesc(&techDesc);


	//Ƽ
	
	//
	XMMATRIX world = XMLoadFloat4x4(&m_gridWorld);
	
	m_pFxWorldViewProj->SetMatrix(reinterpret_cast<float*>(&(world*view*proj)));
	for (UINT i = 0; i < techDesc.Passes; ++i)
	{
		m_pTechnique->GetPassByIndex(i)->Apply(0, m_pImmediateContext);
		m_pImmediateContext->DrawIndexed(m_grid.indices.size(), m_gridIStart, m_gridVStart);
	}

	//
	world = XMLoadFloat4x4(&m_boxWorld);
	m_pFxWorldViewProj->SetMatrix(reinterpret_cast<float*>(&(world*view*proj)));
	for (UINT i = 0; i < techDesc.Passes; ++i)
	{
		m_pTechnique->GetPassByIndex(i)->Apply(0, m_pImmediateContext);
		m_pImmediateContext->DrawIndexed(m_box.indices.size(), m_boxIStart, m_boxVStart);
	}

	//λĵ 5
	world = XMLoadFloat4x4(&m_sphereWorld[4]);
	m_pFxWorldViewProj->SetMatrix(reinterpret_cast<float*>(&(world*view*proj)));
	for (UINT i = 0; i < techDesc.Passes; ++i)
	{
		m_pTechnique->GetPassByIndex(i)->Apply(0, m_pImmediateContext);
		m_pImmediateContext->DrawIndexed(m_sphere.indices.size(), m_sphereIStart, m_sphereVStart);
	}

	//4Բ4
	for (UINT i = 0; i < 4; ++i)
	{
		world = XMLoadFloat4x4(&m_cylinderWorld[i]);
		m_pFxWorldViewProj->SetMatrix(reinterpret_cast<float*>(&(world*view*proj)));
		for (UINT j = 0; j < techDesc.Passes; ++j)
		{
			m_pTechnique->GetPassByIndex(j)->Apply(0, m_pImmediateContext);
			m_pImmediateContext->DrawIndexed(m_cylinder.indices.size(), m_cylinderIStart, m_cylinderVStart);
		}

		world = XMLoadFloat4x4(&m_sphereWorld[i]);
		m_pFxWorldViewProj->SetMatrix(reinterpret_cast<float*>(&(world*view*proj)));
		for (UINT j = 0; j < techDesc.Passes; ++j)
		{
			m_pTechnique->GetPassByIndex(j)->Apply(0, m_pImmediateContext);
			m_pImmediateContext->DrawIndexed(m_sphere.indices.size(), m_sphereIStart, m_sphereVStart);
		}
	}

	m_pSwapChain->Present(0, 0);
}

void GeoDrawDemo::OnMouseDown(WPARAM btnState, int x, int y)
{
	m_lastMousePos.x = x;
	m_lastMousePos.y = y;
	SetCapture(m_hWnd);
}

void GeoDrawDemo::OnMouseUp(WPARAM btnState, int x, int y)
{
	ReleaseCapture();
}

//numberlowhigh֮ 
template<typename T>
static T Clamp(const T& x, const T& low, const T& high)
{
	return x < low ? low : (x > high ? high : x);
}

void GeoDrawDemo::OnMouseMove(WPARAM btnState, int x, int y)
{
	if ((btnState & MK_LBUTTON) != 0)
	{
		//Ƕת
		float dx = XMConvertToRadians(0.25f*static_cast<float>(x - m_lastMousePos.x));
		float dy = XMConvertToRadians(0.25f*static_cast<float>(y - m_lastMousePos.y));

		m_theta += dx;
		m_phi += dy;

		m_phi = Clamp(m_phi, 0.1f, XM_PI - 0.1f);
	}
	else if ((btnState & MK_RBUTTON) != 0)
	{
		float dx = 0.2f*static_cast<float>(x - m_lastMousePos.x);
		float dy = 0.2f*static_cast<float>(y - m_lastMousePos.y);

		m_radius += dx - dy;

		m_radius = Clamp(m_radius, 3.0f, 300.0f);
	}

	m_lastMousePos.x = x;
	m_lastMousePos.y = y;
}


